#ifndef _rtp_BASIS_HH
#define _rtp_BASIS_HH


#include "rtp/namespace.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> class basis;

RTP_NAMESPACE_END

#include "rtp/frame.hh"
#include "rtp/vector.hh"
#include "rtp/mat.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> 
class basis
{
public:
  typedef basis<T,N> this_t;
  typedef mat<T,N,N> rep_t;
  typedef frame<T,N> frame_t;
  typedef vector<T,N> vector_t;

private:
  rep_t _rep;

public:

  basis<T,N>()
  { }

  explicit basis<T,N>(const T &x, const frame_t *f = 0) 
    : _rep(x) 
  { to_frame(f, *this); }

  basis<T,N>(const T *x, const frame_t *f = 0) 
    : _rep(x)
  { to_frame(f, *this); }

  basis<T,N>(const rep_t &x, const frame_t *f = 0) 
    : _rep(x)
  { to_frame(f, *this); }

  basis<T,N>(const T &x1, const T &x2, const frame<T,N> *f = 0) 
    : _rep(x1, x2) 
  { to_frame(f, *this); }

  basis<T,N>(const T &x1, const T &x2, const T &x3, const frame<T,N> *f = 0) 
    : _rep(x1, x2, x3) 
  { to_frame(f, *this); }

  T *operator[](int i) { return _rep[i]; };
  const T *operator[](int i) const { return _rep[i]; };

  const rep_t &rep() const { return _rep; }
  rep_t rep(const frame_t *f) const
  {
    this_t tmp(*this);
    to_frame((frame_t*) 0, tmp, f);
    return tmp.rep();
  }

  // Basis * Scalar => Basis
  this_t &operator *=(const T &x)
  { _rep *= x; return *this; }
  this_t operator *(const T &x)
  { this_t t(*this); return t *= x; }

  // cross product: Basis . Vector => Vector
  this_t    operator *  (const vector_t &x) const
  { return this_t(x.rep() * _rep); }

  // subtraction: Basis + Basis => Basis
  this_t &operator +=(const this_t &x)
  { _rep += x._rep; return *this; }
  this_t operator +(const this_t &x) const
  { this_t t(*this); return t += x; }

  // subtraction: Basis - Basis => Basis
  this_t &operator -=(const this_t &x)
  { _rep -= x._rep; return *this; }
  this_t operator -(const this_t &x) const
  { this_t t(*this); return t += x; }
  
  // negation: - Basis => Basis
  this_t &negate()
  { _rep.negate(); return *this; }
  this_t operator -() const
  { this_t t(*this); t.negate(); return t; }


};


template <class OS, class T, int N> 
OS &operator <<(OS &os, const basis<T,N> &x)
{
  os << "basis " << x.rep();

  return os;
}

RTP_NAMESPACE_END


#endif
