#ifndef _rtp_FRAME_HH
#define _rtp_FRAME_HH

#include "rtp/namespace.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> class frame;
template <class T, int N, class X>
void to_frame(const frame<T,N> *from, X &x, const frame<T,N> *to = 0);

RTP_NAMESPACE_END

#include "location.hh"
#include "vector.hh"
#include "orientation.hh"


RTP_NAMESPACE_BEGIN

template <class T, int N>
class frame
{
public:
  typedef frame<T,N> this_t;
  typedef vector<T,N> vector_t;
  typedef orientation<T,N> orientation_t;
  typedef location<T,N> location_t;

  static const this_t *lab() { return (const this_t *) 0; }

protected:
  location_t _p;    //@ origin offset from parent frame.
  orientation_t _r; //@ orientation around local origin(rotation)

public:
  frame(const location_t &P, const orientation_t &R) 
    : _p(P), _r(R)
  {
  }

  
  ///////////////////////////////////////////////////////////
  // Accessors.

  const location_t &origin() const { return _p; }
  const orientation_t &orientation() const { return _r; }


  ///////////////////////////////////////////////////////////
  // vector

  void to_lab(vector_t &x) const
  {
    // Apply rotation to lab frame.
    x = x * _r;
  }


  void from_lab(vector_t &x) const
  {
    // Apply rotation from lab frame.
    x = x / _r;
  }


  ///////////////////////////////////////////////////////////
  // orientation


  void to_lab(orientation_t &x) const
  {
    // Apply rotation to lab frame.
    x = x * _r;
  }


  void from_lab(orientation_t &x) const
  {
    // Apply rotation from lab frame.
    x = x / _r;
  }


  ///////////////////////////////////////////////////////////
  // location


  void to_lab(location_t &x) const
  {
    // Rotate to parent frame.
    // Translate to parent frame.
    x = (x * _r).rep() + _p.rep();
  }


  void from_lab(location_t &x) const
  {
    // Translate from parent frame.
    // Rotate from parent frame.
    x = ((x - _p) / _r).rep();
  }
};


template <class T, int N, class X>
void to_frame(const frame<T,N> *from, X &x, const frame<T,N> *to = 0)
{
  if ( from == to ) return;
  if ( from != 0 ) from->to_lab(x);
  if ( to != 0 )   to->from_lab(x);
}


template <class OS, class T, int N> 
OS &operator <<(OS &os, const frame<T,N> &x)
{
  os << "frame {\n  " 
     << x.origin() << ",\n  "
     << x.orientation() << "\n}";

  return os;
}


RTP_NAMESPACE_END

#endif
