#ifndef _rtp_LOCATION_HH
#define _rtp_LOCATION_HH

#include "rtp/namespace.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> class location;

RTP_NAMESPACE_END

#include "rtp/frame.hh"
#include "rtp/vector.hh"
#include "rtp/vec.hh"

RTP_NAMESPACE_BEGIN

template <class T, int N> 
class location
{
public:
  typedef location<T,N> this_t;
  typedef location<T,N> location_t;
  typedef vec<T,N> rep_t;
  typedef frame<T,N> frame_t;
  typedef vector<T,N> vector_t;

  friend class vector<T,N>;
  rep_t _v;

public:
  location<T,N>() { }

  location<T,N>(const T &x, const frame_t *f = 0) 
    : _v(x) 
  { to_frame(f, *this); }

  location<T,N>(const rep_t &x, const frame_t *f = 0) 
    : _v(x) 
  { to_frame(f, *this); }

  location<T,N>(const T &_0, const T &_1, const frame_t *f = 0) 
    : _v(_0, _1) 
  { to_frame(f, *this); }

  location<T,N>(const T &_0, const T &_1, const T &_2, const frame_t *f = 0) 
    : _v(_0, _1, _2) 
  { to_frame(f, *this); }


  const rep_t &rep() const { return _v; }
  rep_t rep(const frame_t *f) const
  {
    this_t tmp(*this);
    to_frame((frame_t *) 0, tmp, f);
    return tmp.rep();
  }

  T &operator[](int i) { return _v[i]; };
  const T &operator[](int i) const { return _v[i]; };

  // displacement: Location + Vector => Location
  location_t& operator +=(const vector_t &v)
  { _v += v.rep(); return *this; }
  location_t operator +(const vector_t &v) const
  { location_t tmp(*this); tmp += v; return tmp; }

  // displacement: Location - Vector => Location
  location_t& operator -=(const vector_t &v)
  { _v -= v.rep(); return *this; }
  location_t operator -(const vector_t &v) const
  { location_t tmp(*this); tmp -= v; return tmp; }

  // difference: Location - Location => Vector
  vector_t operator -(const location_t &v) const
  { return vector_t(_v - v.rep()); }
};


template <class OS, class T, int N> 
OS &operator <<(OS &os, const location<T,N> &x)
{
  os << "location " << x.rep();

  return os;
}

RTP_NAMESPACE_END

#endif
