#ifndef _rtp_state_hh
#define _rtp_state_hh

#include <vector>
#include <list>

#include "rtp/namespace.hh"

RTP_NAMESPACE_BEGIN

template class State<T> {
  typedef vector<T> _v_t;
  typedef list<_v_t> _lv_t;

  enum {
    MAX_LEN = 512;
  };

public:
  class StateIterator<T> {
    _lv_t *_l;
    _lv_t::iterator _l_i, _l_end;
    _v_t *_v;
    _v_t::iterator _v_i, _v_end;

  public:
    void reset() 
    {
      _l_i = _l->begin();
      _l_end = _l->end();
      _v = 0;
      _v_i = (*_l_i).begin();
      _v_end = (*_l_i).end();
    }

    StateIterator<T>(State<T> *l) 
      : _l(l) 
    { 
    }

    template <class I> void put(I x, I x_end)
    {
      while ( x != x_end) {
	if ( _v == 0 || _v->size() >= MAX_LEN ) {
	  _l->push_back();
	  _l_i = (_l_end = _l->end());
	  _v = &(*(++ _l_i));
	  _v->size(MAX_LEN);
	  _v_i = _v->begin();
	  _v_end = _v->end();
	}
	_v->insert(*(x ++));
      }
    }

    template <class I> void get(I x, I x_end)
    {
      while ( x != x_end ) {
	*(x ++) = *(_v_i ++);
	if ( _v_i == _v_end ) {
	  if ( ++ _l_i != _i_end ) {
	    _v = &(*(_l_i));
	    _v_i = _v->begin();
	    _v_end = _v->end();
	  } else {
	    abort();
	  }
	}
      }
    }

  };

  typedef StateIterator<T> iterator;

private:
  //////////////////////////////////////

  friend class StateIterator<T>;

  _lv_t _l;
  size_t _n;
  iterator _i;

  //////////////////////////////////////

  void init()
  {
    _l.push_back();
    _v->size(MAX_LEN);
    _i._l = this;
    _i.reset();
  }

public:
  //////////////////////////////////////

  State<T> 
  : _l(),
    _n(0),
    _i(this)
  {
    init();
  }
  ~State<T> { }
  
  void clear()
  {
    _l.clear();
    _n = 0;
  }

};


template <class I> void get(I x, I x_end) { _i.get(x, x_end); }
template <class I> void put(I x, I x_end) { _i.put(x, x_end); }

template <class I, class T> void put(I &x, const T &x) { _i.put(&x, 1); }
template <class I, class T> void get(I &x, const T &x) { _i.get(&x, 1); }

template <class S, class T> State<S>& operator <<(State<S>& s, const T &x)
{
  get(s, x);
  return s;
}


template <class S, class T> State<S>& operator >>(State<S>& s, const T &x)
{
  put(s, x);
  return s;
}


RTP_NAMESPACE_END


#endif
