#ifndef _rtp_VEC_HH
#define _rtp_VEC_HH


#include "rtp/namespace.hh"

#include <assert.h>
#include <math.h> /* sqrt() */

#define _rtp_UNROLL(N,I,E) \

RTP_NAMESPACE_BEGIN

template <class T, int N> 
class vec {
public:
  typedef vec<T,N> this_t;
  typedef vec<T,N> vec_t;

protected:
  T _rep[N];

  static void init(T *v, const T &x)
  {
    v[0] = x;
    if ( N > 1 ) v[1] = x;
    if ( N > 2 ) v[2] = x;
    if ( N > 3 ) v[3] = x;
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	v[i] = x;
      }
    }
  }

  static void init(T *v, const T *x)
  {
    v[0] = x[0];
    if ( N > 1 ) v[1] = x[1];
    if ( N > 2 ) v[2] = x[2];
    if ( N > 3 ) v[3] = x[3];
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	v[i] = x[i];
      }
    }
  }

public:
  vec<T,N>() { };

  explicit vec<T,N>(const T &x)
  { init(_rep, x); }

  vec<T,N>(const T &_0, const T &_1)
  { assert(N == 2); _rep[0] = _0; _rep[1] = _1; };

  vec<T,N>(const T &_0, const T &_1, const T &_2)
  { assert(N == 3); _rep[0] = _0; _rep[1] = _1; _rep[2] = _2; };

  vec<T,N>(const T &_0, const T &_1, const T &_2, const T &_3)
  { assert(N == 4); _rep[0] = _0; _rep[1] = _1; _rep[2] = _2; _rep[3] = _3; };

  vec<T,N>(const T *x) { init(_rep, x); }
  vec<T,N>(const vec<T,N> &x) { init(_rep, x._rep); }


  T &operator[](unsigned int i) 
  { 
    assert(i <  N); 
    return _rep[i];
  }
  const T &operator[](unsigned int i) const 
  { 
    assert(i <  N); 
    return _rep[i];
  }


  vec_t &operator +=(const vec_t &x) //@ Addition.
  {
    _rep[0] += x._rep[0];
    if ( N > 1 ) _rep[1] += x._rep[1];
    if ( N > 2 ) _rep[2] += x._rep[2];
    if ( N > 3 ) _rep[3] += x._rep[3];
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	_rep[i] += x._rep[i];
      }
    }
    return *this;
  }
  vec_t operator +(const vec_t &x) const 
  { vec_t t(*this); return t += x; }


  vec_t &operator -=(const vec_t &x) //@ Subtraction.
  { 
    _rep[0] -= x._rep[0];
    if ( N > 1 ) _rep[1] -= x._rep[1];
    if ( N > 2 ) _rep[2] -= x._rep[2];
    if ( N > 3 ) _rep[3] -= x._rep[3];
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	_rep[i] -= x._rep[i];
      }
    }
     return *this;
  }
  vec_t operator -(const vec_t &x) const
  { vec_t t(*this); return t -= x; }


  vec_t &operator *=(const T &x) //@ Scale.
  { 
    _rep[0] *= x;
    if ( N > 1 ) _rep[1] *= x;
    if ( N > 2 ) _rep[2] *= x;
    if ( N > 3 ) _rep[3] *= x;
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	_rep[i] *= x;
      }
    }
    return *this;
  }
  vec_t operator *(const T &x) const
  { vec_t t(*this); return t *= x; }


  vec_t &operator /=(const T &x) //@ Scale.
  { 
    _rep[0] /= x;
    if ( N > 1 ) _rep[1] /= x;
    if ( N > 2 ) _rep[2] /= x;
    if ( N > 3 ) _rep[3] /= x;
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	rep[i] /= x;
      }
    }
    return *this;
  }
  vec_t operator /(const T &x) const
  { vec_t t(*this); return t /= x; }


   // inner product : Vector . Vector -> Scalars
  T operator %(const vec_t &x) const //@ Dot product.
  {
    T p = _rep[0] * x._rep[0];
    if ( N > 1 ) p += _rep[1] * x._rep[1];
    if ( N > 2 ) p += _rep[2] * x._rep[2];
    if ( N > 3 ) p += _rep[3] * x._rep[3];
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	p += _rep[i] * x._rep[i];
      }
    }
    return p;
  }

  // Cross product: Vector ^ Vector -> Vector
  vec_t operator ^(const vec_t &x) const 
  { 
    assert(N == 3);
    return vec_t(
		 _rep[1] * x._rep[2] - _rep[2] * x._rep[1],
		 _rep[2] * x._rep[0] - _rep[0] * x._rep[2],
		 _rep[0] * x._rep[1] - _rep[1] * x._rep[0]
		 );
  }
  vec_t &operator ^=(const vec_t &x)
  { (*this) = (*this) ^ x; return *this; }


  // Outer product: Vector . Vector -> Vector
  vec_t &operator *=(const vec_t &x)
  { 
    _rep[0] *= x._rep[0];
    if ( N > 1 ) _rep[1] *= x._rep[1];
    if ( N > 2 ) _rep[2] *= x._rep[2];
    if ( N > 3 ) _rep[3] *= x._rep[3];
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	_rep[i] *= x._rep[i];
      }
    }
    return *this;
  }
  vec_t operator *(const vec_t &x) const 
  { vec_t t(*this); return t *= x; }

  
  vec_t &operator /=(const vec_t &x)
  { 
    _rep[0] /= x._rep[0];
    if ( N > 1 ) _rep[1] /= x._rep[1];
    if ( N > 2 ) _rep[2] /= x._rep[2];
    if ( N > 3 ) _rep[3] /= x._rep[3];
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	_rep[i] /= x._rep[i];
      }
    }
    return *this;
  }
  vec_t operator /(const vec_t &x) const 
  { vec_t t(*this); return t /= x; }

  
  // negation: - Vector => Vector
  vec_t &negate()
  {
    _rep[0] = - _rep[0];
    if ( N > 1 ) _rep[1] = - _rep[1];
    if ( N > 2 ) _rep[2] = - _rep[2];
    if ( N > 3 ) _rep[3] = - _rep[3];
    if ( N > 4 ) {
      for ( unsigned int i = 4; i < N; ++ i ) {
	rep[i] = - _rep[i];
      }
    }
    return *this;
  }
  vec_t operator -() const
  { vec_t t(*this); t.negate(); return t; }

  // Norm: | Vector | => Scalar
  T norm() const
  { return sqrt(*this % *this); }
  operator T() const //@ Norm
  { return norm(); }

  // Normal vector: || Vector || => Vector
  vec_t &normalize()
  {
    T n = norm();
    if ( n ) {
      n = 1 / n;
      (*this) *= n;
    }
    return *this;
  }
  vec_t operator !() const
  { vec_t t(*this); return t.normalize(); }

  //@ Orthogonal vector
  vec_t operator ~() const
  {
    if ( N == 2 ) {
      // 1,0 => 0,1
      return vec_t(- _rep[1], _rep[0]);
    }
    if ( N == 3 ) {
      // CHECK THIS CODE!!!.
      // 1,0,0 => 0,1,0
      // 0,1,0 => 0,0,1
      // 0,0,1 => 1,0,0
      return vec_t(- _rep[2], _rep[0], _rep[1]);
    }
  }

};


template <class T, int N>
vec<T,N> operator * ( const T &x, const vec<T,N> &v )
{
  return v * x;
}


template <class OS, class T, int N> 
OS &operator << ( OS &os, const vec<T,N> &x )
{
  os << "{ " << x[0];
  if ( N > 1 ) os << ", " << x[1];
  if ( N > 2 ) os << ", " << x[2];
  if ( N > 3 ) os << ", " << x[3];
  if ( N > 4 ) {
    for ( unsigned int i = 4; i < N; ++ i ) {
      os << ", " << x[i];
    }
  }
  os << " }";

  return os;
}


template <class OS, class T, int N> 
OS &operator >> ( OS &os, vec<T,N> &x )
{
  char c;

  os >> c;
  os >> x[0];
  if ( N > 1 ) { os >> c; os >> x[1]; }
  if ( N > 2 ) { os >> c; os >> x[2]; }
  if ( N > 3 ) { os >> c; os >> x[3]; }
  if ( N > 4 ) {
    for ( unsigned int i = 4; i < N; ++ i ) {
      os >> c; os >> x[i];
    }
  }
  os >> c;

  return os;
}

RTP_NAMESPACE_END

#endif
